/*
  Copyright (C) 2011 - 2019 by the authors of the ASPECT code.

  This file is part of ASPECT.

  ASPECT is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  ASPECT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ASPECT; see the file LICENSE.  If not see
  <http://www.gnu.org/licenses/>.
*/


#include <aspect/material_model/latent_heat.h>
#include <aspect/adiabatic_conditions/interface.h>
#include <aspect/gravity_model/interface.h>


namespace aspect
{
  namespace MaterialModel
  {
    template <int dim>
    double
    LatentHeat<dim>::
    phase_function (const Point<dim> &position,
                    const double temperature,
                    const double pressure,
                    const int phase) const
    {
      // We need to convert the depth to pressure (depth-based phase transitions),
      // or else use the pressure itself if the phase transition is defined based off a pressure.
      if (!use_depth)
        {
          // first, get the pressure at which the phase transition occurs normally
          // and get the pressure change in the range of the phase transition

          double transition_pressure;
          double pressure_width;
          double width_temp;

          transition_pressure = transition_pressures[phase];
          pressure_width = transition_pressure_widths[phase];
          width_temp = transition_pressure_widths[phase];

          // then calculate the deviation from the transition point (both in temperature
          // and in pressure)
          double pressure_deviation = pressure - transition_pressure
                                      - transition_slopes[phase] * (temperature - transition_temperatures[phase]);

          // last, calculate the percentage of material that has undergone the transition
          // (also in dependence of the phase transition width - this is an input parameter)
          double phase_func;
          // use delta function for width = 0
          if (width_temp==0)
            (pressure_deviation > 0) ? phase_func = 1 : phase_func = 0;
          else
            phase_func = 0.5*(1.0 + std::tanh(pressure_deviation / pressure_width));
          return phase_func;
        }
      // this part of the loop is only implemented for phase transitions based off of depth,
      // since pressure-based transitions are included above.
      else
        {
          double depth = this->get_geometry_model().depth(position);
          double depth_deviation = (pressure > 0
                                    ?
                                    depth - transition_depths[phase]
                                    - transition_slopes[phase] * (depth / pressure) * (temperature - transition_temperatures[phase])
                                    :
                                    depth - transition_depths[phase]
                                    - transition_slopes[phase] / (this->get_gravity_model().gravity_vector(position).norm() * reference_rho)
                                    * (temperature - transition_temperatures[phase]));
          double phase_func;
          // use delta function for width = 0
          if (transition_widths[phase]==0)
            phase_func = (depth_deviation > 0) ? 1 : 0;
          else
            phase_func = 0.5*(1.0 + std::tanh(depth_deviation / transition_widths[phase]));
          return phase_func;
        }
    }

    template <int dim>
    double
    LatentHeat<dim>::
    phase_function_derivative (const Point<dim> &,
                               const double temperature,
                               const double pressure,
                               const int phase) const
    {
      double transition_pressure;
      double pressure_width;
      double width_temp;

      // we already should have the adiabatic conditions here
      AssertThrow (this->get_adiabatic_conditions().is_initialized(),
                   ExcMessage("need adiabatic conditions to incorporate phase transitions"));

      // first, get the pressure at which the phase transition occurs normally

      // phase transition based off of depth
      if (use_depth)
        {
          const Point<dim,double> transition_point = this->get_geometry_model().representative_point(transition_depths[phase]);
          const Point<dim,double> transition_plus_width = this->get_geometry_model().representative_point(transition_depths[phase] + transition_widths[phase]);
          const Point<dim,double> transition_minus_width = this->get_geometry_model().representative_point(transition_depths[phase] - transition_widths[phase]);
          transition_pressure = this->get_adiabatic_conditions().pressure(transition_point);
          pressure_width = 0.5 * (this->get_adiabatic_conditions().pressure(transition_plus_width)
                                  - this->get_adiabatic_conditions().pressure(transition_minus_width));
          width_temp = transition_widths[phase];
        }
      // using pressure instead of depth to define the phase transition
      else
        {
          transition_pressure = transition_pressures[phase];
          pressure_width = transition_pressure_widths[phase];
          width_temp = transition_pressure_widths[phase];
        }

      // then calculate the deviation from the transition point (both in temperature
      // and in pressure)
      double pressure_deviation = pressure - transition_pressure
                                  - transition_slopes[phase] * (temperature - transition_temperatures[phase]);

      // last, calculate the analytical derivative of the phase function
      if (width_temp==0)
        return 0;
      else
        return 0.5 / pressure_width * (1.0 - std::tanh(pressure_deviation / pressure_width)
                                       * std::tanh(pressure_deviation / pressure_width));
    }

    template <int dim>
    void
    LatentHeat<dim>::
    evaluate(const MaterialModelInputs<dim> &in,
             MaterialModelOutputs<dim> &out) const
    {
      for (unsigned int i=0; i < in.position.size(); ++i)
        {
          const double temperature = in.temperature[i];
          const double pressure = in.pressure[i];
          const std::vector<double> composition = in.composition[i];
          const Point<dim> position = in.position[i];

          // Assign constant material properties
          {
            out.specific_heat[i] = reference_specific_heat;
            out.thermal_conductivities[i] = k_value;
            out.thermal_expansion_coefficients[i] = thermal_alpha;
            out.compressibilities[i] = reference_compressibility;
          }

          // Calculate Viscosity
          {
            const double delta_temp = temperature-reference_T;
            double visc_temperature_dependence = std::max(std::min(std::exp(-thermal_viscosity_exponent*delta_temp/reference_T),1e2),1e-2);

            if (std::isnan(visc_temperature_dependence))
              visc_temperature_dependence = 1.0;

            double visc_composition_dependence = 1.0;
            if ((composition_viscosity_prefactor != 1.0) && (composition.size() > 0))
              {
                // geometric interpolation
                out.viscosities[i] = (pow(10, ((1-composition[0]) * log10(eta*visc_temperature_dependence)
                                               + composition[0] * log10(eta*composition_viscosity_prefactor*visc_temperature_dependence))));
              }
            else
              out.viscosities[i] = visc_composition_dependence * visc_temperature_dependence * eta;
          }

          // Calculate density
          // and phase dependence of viscosity
          {
            // first, calculate temperature dependence of density
            double density_temperature_dependence = 1.0;
            if (this->include_adiabatic_heating ())
              {
                // temperature dependence is 1 - alpha * (T - T(adiabatic))
                density_temperature_dependence -= (temperature - this->get_adiabatic_conditions().temperature(position))
                                                  * thermal_alpha;
              }
            else
              density_temperature_dependence -= temperature * thermal_alpha;

            // second, calculate composition dependence of density
            // constant density difference between peridotite and eclogite
            const double density_composition_dependence = composition.size()>0
                                                          ?
                                                          compositional_delta_rho * composition[0]
                                                          :
                                                          0.0;

            // third, calculate the density (and viscosity) differences due to phase
            // transitions (temperature- and pressure dependence included).
            // the phase function gives the percentage of material that has
            // already undergone the phase transition to the higher-pressure material
            // (this is done individual for each transitions and summed up
            // in the end)
            // this means, that there are no actual density or viscosity "jumps", but
            // gradual transitions between the materials
            double phase_dependence = 0.0;
            double viscosity_phase_dependence = 1.0;

            // transitions defined by depth
            unsigned int number_of_phase_transitions;
            if (use_depth)
              number_of_phase_transitions = transition_depths.size();
            // transitions defined by pressure
            else
              number_of_phase_transitions = transition_pressures.size();

            // note that for the densities, we have a list of jumps, so the index used
            // in the loop corresponds to the index of the phase transition, whereas
            // for the viscosities we have a list of prefactors (which has one more
            // entry for the first layer), so we have to use i+1 as index
            if (composition.size()==0)      // only one field
              {
                for (unsigned int phase=0; phase<number_of_phase_transitions; ++phase)
                  {
                    const double phaseFunction = phase_function (position,
                                                                 temperature,
                                                                 pressure,
                                                                 phase);

                    phase_dependence += phaseFunction * density_jumps[phase];
                    viscosity_phase_dependence *= 1. + phaseFunction * (phase_prefactors[phase+1]-1.);
                  }
              }
            else if (composition.size()>0)
              {
                for (unsigned int phase=0; phase<number_of_phase_transitions; ++phase)
                  {
                    const double phaseFunction = phase_function (position,
                                                                 temperature,
                                                                 pressure,
                                                                 phase);
                    if (transition_phases[phase] == 0)     // 1st compositional field
                      phase_dependence += phaseFunction * density_jumps[phase] * (1.0 - composition[0]);
                    else if (transition_phases[phase] == 1) // 2nd compositional field
                      phase_dependence += phaseFunction * density_jumps[phase] * composition[0];
                    viscosity_phase_dependence *= 1. + phaseFunction * (phase_prefactors[phase]-1.);
                  }
              }
            // fourth, pressure dependence of density
            const double kappa = reference_compressibility;
            const double pressure_dependence = reference_rho * kappa * (pressure - this->get_surface_pressure());

            // in the end, all the influences are added up
            out.densities[i] = (reference_rho + density_composition_dependence + pressure_dependence + phase_dependence)
                               * density_temperature_dependence;
            out.viscosities[i] = std::max(min_viscosity, std::min(max_viscosity, out.viscosities[i] * viscosity_phase_dependence));
          }

          // Calculate entropy derivative
          {
            double entropy_gradient_pressure = 0.0;
            double entropy_gradient_temperature = 0.0;
            const double rho = out.densities[i];
            unsigned int number_of_phase_transitions;

            // transition defined by depth
            if (use_depth)
              number_of_phase_transitions= transition_depths.size();
            // transition defined by pressure
            else
              number_of_phase_transitions= transition_pressures.size();


            if (this->get_adiabatic_conditions().is_initialized() && this->include_latent_heat())
              for (unsigned int phase=0; phase<number_of_phase_transitions; ++phase)
                {
                  // calculate derivative of the phase function
                  const double PhaseFunctionDerivative = phase_function_derivative(position,
                                                                                   temperature,
                                                                                   pressure,
                                                                                   phase);

                  // calculate the change of entropy across the phase transition
                  double entropy_change = 0.0;
                  if (composition.size()==0)      // only one compositional field
                    entropy_change = transition_slopes[phase] * density_jumps[phase] / (rho * rho);
                  else
                    {
                      if (transition_phases[phase] == 0)     // 1st compositional field
                        entropy_change = transition_slopes[phase] * density_jumps[phase] / (rho * rho) * (1.0 - composition[0]);
                      else if (transition_phases[phase] == 1) // 2nd compositional field
                        entropy_change = transition_slopes[phase] * density_jumps[phase] / (rho * rho) * composition[0];
                    }
                  // we need DeltaS * DX/Dpressure_deviation for the pressure derivative
                  // and - DeltaS * DX/Dpressure_deviation * gamma for the temperature derivative
                  entropy_gradient_pressure += PhaseFunctionDerivative * entropy_change;
                  entropy_gradient_temperature -= PhaseFunctionDerivative * entropy_change * transition_slopes[phase];
                }
            out.entropy_derivative_pressure[i] = entropy_gradient_pressure;
            out.entropy_derivative_temperature[i] = entropy_gradient_temperature;
          }

          // Assign reaction terms
          for (unsigned int c=0; c<in.composition[i].size(); ++c)
            out.reaction_terms[i][c] = 0.0;

        }
    }


    template <int dim>
    double
    LatentHeat<dim>::
    reference_viscosity () const
    {
      return eta;
    }




    template <int dim>
    bool
    LatentHeat<dim>::
    is_compressible () const
    {
      if (reference_compressibility > 0)
        return true;
      else
        return false;
    }



    template <int dim>
    void
    LatentHeat<dim>::declare_parameters (ParameterHandler &prm)
    {
      prm.enter_subsection("Material model");
      {
        prm.enter_subsection("Latent heat");
        {
          prm.declare_entry ("Reference density", "3300",
                             Patterns::Double (0),
                             "Reference density $\\rho_0$. Units: $kg/m^3$.");
          prm.declare_entry ("Reference temperature", "293",
                             Patterns::Double (0),
                             "The reference temperature $T_0$. Units: $K$.");
          prm.declare_entry ("Viscosity", "5e24",
                             Patterns::Double (0),
                             "The value of the constant viscosity. Units: $kg/m/s$.");
          prm.declare_entry ("Composition viscosity prefactor", "1.0",
                             Patterns::Double (0),
                             "A linear dependency of viscosity on composition. Dimensionless prefactor.");
          prm.declare_entry ("Thermal viscosity exponent", "0.0",
                             Patterns::Double (0),
                             "The temperature dependence of viscosity. Dimensionless exponent.");
          prm.declare_entry ("Thermal conductivity", "2.38",
                             Patterns::Double (0),
                             "The value of the thermal conductivity $k$. "
                             "Units: $W/m/K$.");
          prm.declare_entry ("Reference specific heat", "1250",
                             Patterns::Double (0),
                             "The value of the specific heat $C_p$. "
                             "Units: $J/kg/K$.");
          prm.declare_entry ("Thermal expansion coefficient", "4e-5",
                             Patterns::Double (0),
                             "The value of the thermal expansion coefficient $\\beta$. "
                             "Units: $1/K$.");
          prm.declare_entry ("Compressibility", "5.124e-12",
                             Patterns::Double (0),
                             "The value of the compressibility $\\kappa$. "
                             "Units: $1/Pa$.");
          prm.declare_entry ("Density differential for compositional field 1", "0",
                             Patterns::Double(),
                             "If compositional fields are used, then one would frequently want "
                             "to make the density depend on these fields. In this simple material "
                             "model, we make the following assumptions: if no compositional fields "
                             "are used in the current simulation, then the density is simply the usual "
                             "one with its linear dependence on the temperature. If there are compositional "
                             "fields, then the density only depends on the first one in such a way that "
                             "the density has an additional term of the kind $+\\Delta \\rho \\; c_1(\\mathbf x)$. "
                             "This parameter describes the value of $\\Delta \\rho$. Units: $kg/m^3/\\textrm{unit "
                             "change in composition}$.");
          prm.declare_entry ("Phase transition depths", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of depths where phase transitions occur. Values must "
                             "monotonically increase. "
                             "Units: $m$.");
          prm.declare_entry ("Phase transition widths", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of widths for each phase transition, in terms of depth. The phase functions "
                             "are scaled with these values, leading to a jump between phases "
                             "for a value of zero and a gradual transition for larger values. "
                             "List must have the same number of entries as Phase transition depths. "
                             "Units: $m$.");
          prm.declare_entry ("Phase transition pressures", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of pressures where phase transitions occur. Values must "
                             "monotonically increase. Define transition by depth instead of "
                             "pressure must be set to false to use this parameter. "
                             "Units: $Pa$.");
          prm.declare_entry ("Phase transition pressure widths", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of widths for each phase transition, in terms of pressure. The phase functions "
                             "are scaled with these values, leading to a jump between phases "
                             "for a value of zero and a gradual transition for larger values. "
                             "List must have the same number of entries as Phase transition pressures. "
                             "Define transition by depth instead of pressure must be set to false "
                             "to use this parameter. "
                             "Units: $Pa$.");
          prm.declare_entry ("Define transition by depth instead of pressure", "true",
                             Patterns::Bool (),
                             "Whether to list phase transitions by depth or pressure. If this parameter is true, "
                             "then the input file will use Phase transitions depths and Phase transition widths "
                             "to define the phase transition. If it is false, the parameter file will read in "
                             "phase transition data from Phase transition pressures and "
                             "Phase transition pressure widths.");
          prm.declare_entry ("Phase transition temperatures", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of temperatures where phase transitions occur. Higher or lower "
                             "temperatures lead to phase transition occurring in smaller or greater "
                             "depths than given in Phase transition depths, depending on the "
                             "Clapeyron slope given in Phase transition Clapeyron slopes. "
                             "List must have the same number of entries as Phase transition depths. "
                             "Units: $K$.");
          prm.declare_entry ("Phase transition Clapeyron slopes", "",
                             Patterns::List (Patterns::Double()),
                             "A list of Clapeyron slopes for each phase transition. A positive "
                             "Clapeyron slope indicates that the phase transition will occur in "
                             "a greater depth, if the temperature is higher than the one given in "
                             "Phase transition temperatures and in a smaller depth, if the "
                             "temperature is smaller than the one given in Phase transition temperatures. "
                             "For negative slopes the other way round. "
                             "List must have the same number of entries as Phase transition depths. "
                             "Units: $Pa/K$.");
          prm.declare_entry ("Phase transition density jumps", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of density jumps at each phase transition. A positive value means "
                             "that the density increases with depth. The corresponding entry in "
                             "Corresponding phase for density jump determines if the density jump occurs "
                             "in peridotite, eclogite or none of them."
                             "List must have the same number of entries as Phase transition depths. "
                             "Units: $kg/m^3$.");
          prm.declare_entry ("Corresponding phase for density jump", "",
                             Patterns::List (Patterns::Integer(0)),
                             "A list of phases, which correspond to the Phase transition density jumps. "
                             "The density jumps occur only in the phase that is given by this phase value. "
                             "0 stands for the 1st compositional fields, 1 for the second compositional field "
                             "and -1 for none of them. "
                             "List must have the same number of entries as Phase transition depths. "
                             "Units: $Pa/K$.");
          prm.declare_entry ("Viscosity prefactors", "",
                             Patterns::List (Patterns::Double(0)),
                             "A list of prefactors for the viscosity for each phase. The reference "
                             "viscosity will be multiplied by this factor to get the corresponding "
                             "viscosity for each phase. "
                             "List must have one more entry than Phase transition depths. "
                             "Units: non-dimensional.");
          prm.declare_entry ("Minimum viscosity", "1e19",
                             Patterns::Double (0),
                             "Limit for the minimum viscosity in the model. "
                             "Units: Pa \\, s.");
          prm.declare_entry ("Maximum viscosity", "1e24",
                             Patterns::Double (0),
                             "Limit for the maximum viscosity in the model. "
                             "Units: Pa \\, s.");
        }
        prm.leave_subsection();
      }
      prm.leave_subsection();
    }



    template <int dim>
    void
    LatentHeat<dim>::parse_parameters (ParameterHandler &prm)
    {
      prm.enter_subsection("Material model");
      {
        prm.enter_subsection("Latent heat");
        {
          reference_rho              = prm.get_double ("Reference density");
          reference_T                = prm.get_double ("Reference temperature");
          eta                        = prm.get_double ("Viscosity");
          composition_viscosity_prefactor = prm.get_double ("Composition viscosity prefactor");
          thermal_viscosity_exponent = prm.get_double ("Thermal viscosity exponent");
          k_value                    = prm.get_double ("Thermal conductivity");
          reference_specific_heat    = prm.get_double ("Reference specific heat");
          thermal_alpha              = prm.get_double ("Thermal expansion coefficient");
          reference_compressibility  = prm.get_double ("Compressibility");
          compositional_delta_rho    = prm.get_double ("Density differential for compositional field 1");
          min_viscosity              = prm.get_double ("Minimum viscosity");
          max_viscosity              = prm.get_double ("Maximum viscosity");


          transition_depths = Utilities::string_to_double
                              (Utilities::split_string_list(prm.get ("Phase transition depths")));
          transition_widths= Utilities::string_to_double
                             (Utilities::split_string_list(prm.get ("Phase transition widths")));
          transition_pressures = Utilities::string_to_double
                                 (Utilities::split_string_list(prm.get ("Phase transition pressures")));
          transition_pressure_widths= Utilities::string_to_double
                                      (Utilities::split_string_list(prm.get ("Phase transition pressure widths")));
          use_depth                   = prm.get_bool ("Define transition by depth instead of pressure");
          transition_temperatures = Utilities::string_to_double
                                    (Utilities::split_string_list(prm.get ("Phase transition temperatures")));
          transition_slopes = Utilities::string_to_double
                              (Utilities::split_string_list(prm.get ("Phase transition Clapeyron slopes")));
          density_jumps = Utilities::string_to_double
                          (Utilities::split_string_list(prm.get ("Phase transition density jumps")));
          transition_phases = Utilities::string_to_int
                              (Utilities::split_string_list(prm.get ("Corresponding phase for density jump")));
          phase_prefactors = Utilities::string_to_double
                             (Utilities::split_string_list(prm.get ("Viscosity prefactors")));

          // make sure to check against the depth lists for size errors, since using depth
          if (use_depth)
            {
              if (transition_widths.size() != transition_depths.size() ||
                  transition_temperatures.size() != transition_depths.size() ||
                  transition_slopes.size() != transition_depths.size() ||
                  density_jumps.size() != transition_depths.size() ||
                  transition_phases.size() != transition_depths.size() ||
                  phase_prefactors.size() != transition_depths.size()+1)
                AssertThrow(false, ExcMessage("Error: At least one list that gives input parameters for the phase "
                                              "transitions has the wrong size. Currently checking against transition depths. "
                                              "If phase transitions in terms of pressure inputs are desired, check to make sure "
                                              "'Define transition by depth instead of pressure = false'."));
            }
          // make sure to check against the pressure lists for size errors,
          // since pressure is being used instead of depth.
          else
            {
              if (transition_pressure_widths.size() != transition_pressures.size() ||
                  transition_temperatures.size() != transition_pressures.size() ||
                  transition_slopes.size() != transition_pressures.size() ||
                  density_jumps.size() != transition_pressures.size() ||
                  transition_phases.size() != transition_pressures.size() ||
                  phase_prefactors.size() != transition_pressures.size()+1)
                AssertThrow(false, ExcMessage("Error: At least one list that gives input parameters for the phase "
                                              "transitions has the wrong size. Currently checking against transition pressures. "
                                              "If phase transitions in terms of depth inputs are desired, check to make sure "
                                              "'Define transition by depth instead of pressure = true'."));
            }

          // as the phase viscosity prefactors are all applied multiplicatively on top of each other,
          // we have to scale them here so that they are relative factors in comparison to the product
          // of the prefactors of all phase above the current one
          for (unsigned int phase=1; phase<phase_prefactors.size(); ++phase)
            {
              phase_prefactors[phase] /= phase_prefactors[phase-1];
            }

          if (thermal_viscosity_exponent!=0.0 && reference_T == 0.0)
            AssertThrow(false, ExcMessage("Error: Material model latent heat with Thermal viscosity exponent can not have reference_T=0."));
        }
        prm.leave_subsection();
      }
      prm.leave_subsection();

      // Declare dependencies on solution variables
      this->model_dependence.viscosity = NonlinearDependence::temperature | NonlinearDependence::compositional_fields;
      this->model_dependence.density = NonlinearDependence::temperature | NonlinearDependence::pressure | NonlinearDependence::compositional_fields;
      this->model_dependence.compressibility = NonlinearDependence::none;
      this->model_dependence.specific_heat = NonlinearDependence::none;
      this->model_dependence.thermal_conductivity = NonlinearDependence::none;
    }
  }
}

// explicit instantiations
namespace aspect
{
  namespace MaterialModel
  {
    ASPECT_REGISTER_MATERIAL_MODEL(LatentHeat,
                                   "latent heat",
                                   "A material model that includes phase transitions "
                                   "and the possibility that latent heat is released "
                                   "or absorbed when material crosses one of the "
                                   "phase transitions of up to two different materials "
                                   "(compositional fields). "
                                   "This model implements a standard approximation "
                                   "of the latent heat terms following Christensen \\& Yuen, 1985. "
                                   "The change of entropy is calculated as "
                                   "$Delta S = \\gamma \\frac{\\Delta\\rho}{\\rho^2}$ with the "
                                   "Clapeyron slope $\\gamma$ and the density change $\\Delta\\rho$ "
                                   "of the phase transition being input parameters. "
                                   "The model employs an analytic phase function in the form "
                                   "$X=\\frac{1}{2} \\left( 1 + \\tanh \\left( \\frac{\\Delta p}{\\Delta p_0} \\right) \\right)$ "
                                   "with $\\Delta p = p - p_{\\text{transition}} - \\gamma \\left( T - T_{\\text{transition}} \\right)$ "
                                   "and $\\Delta p_0$ being the pressure difference over the width "
                                   "of the phase transition (specified as input parameter).")
  }
}
